///////////////////////////
// Socket event handlers //
///////////////////////////

var tbw_listener = {	
	str_w : 0,
	str_r : 0,
	
	onSocketAccepted: function(aServ, aTransport) {		
		// Got a client connection, open read-write streams		
		this.str_w = aTransport.openOutputStream(
			Components.interfaces.nsITransport.OPEN_BLOCKING, 0, 0
		);
		this.str_r =
			aTransport
			.openInputStream(0, 0, 0)
			.QueryInterface(Components.interfaces.nsIAsyncInputStream)
		;
		
		// Assign the reader function (but only on the same thread)		
		var tm = Components.classes["@mozilla.org/thread-manager;1"].getService();
		this.str_r.asyncWait(tbw_reader, 0, 0, tm.mainThread);
	},
	
	onStopListening: function(aServ, aStatus) {
		// Unhandled (that's ok)
	},
	
	write: function(say) {
		this.str_w.write(say, say.length);
		this.str_w.flush();
	}
	
};

var tbw_reader = {
	onInputStreamReady : function(input) {
		
		// Prepare socket stream (allowing for it being closed at the remote end)
		var sin =
			Components.classes["@mozilla.org/scriptableinputstream;1"]
			.createInstance(Components.interfaces.nsIScriptableInputStream)
		;
		try {
			sin.init(input);
			sin.available();
		} catch (ex) {
			return;
		}
		
		// Read all data into buffer
		var request = '';
		while (sin.available()) request += sin.read(512);
		
		// Process message
		tbw_controller.handle(request);
		
		// Keep connection open
		var tm = Components.classes["@mozilla.org/thread-manager;1"].getService();
		input.asyncWait(tbw_reader, 0, 0, tm.mainThread);
	}
};

var tbw_controller = {
	handle : function(input) {
		// Interpret request as HTTP and pass off to the web server
		var response = tbw_http.doRequest(input); // 'Got ' + input.length + ' bytes!\n';
		tbw_listener.write(response);
	}
	
}

/////////////////////////////
// Javascript HTTP handler //
/////////////////////////////

var tbw_http = {
	doRequest : function(input) {
	
		// Validate HTTP request
		var lines = input.split("\n");
		var http_request = lines[0].split(" ");
		
		// Determine action
		switch (http_request[0]) {
			case 'GET': {
				return this.doFile(http_request[1]);
			} break;
			
			default: {
				return this.doError(501, "Unknown request type "+http_request[0]);
			} break;
		}
		
	},

	// Creates an error message. Returns HTTP
	doError : function(code, message) {
		var response = 'HTTP/1.1 '+code+' ';
		switch (code) {
			case 400: { response += "Bad Request"; } break;
			case 403: { response += "Forbidden"; } break;
			case 404: { response += "Not Found"; } break;
			case 500: { response += "Internal Error"; } break;
			case 501: { response += "Not Implemented"; } break;
			default : {
			} break;
		}
		response += "\nContent-Type: text/html\n\n<h1>Error "+code+": "+message+"</h1>";
		return response;
	},
	
	// Retrieves and parses a file. Returns HTTP
	doFile : function(path) {
	
		// Convert path to chrome:// url. TODO: sanitisation
		var q_pos = path.indexOf('?');
		var q_params = '';
		if (q_pos != -1) {
			q_params = path.substr(q_pos+1);
			path = path.substr(0, q_pos);
		}
		var chrome_path = "chrome://thunderbirdwebui/content/www" + path;
		if (path[0] != '/' || chrome_path.indexOf("..") != -1) {
			return this.doError(400, "Potentially malicious request");
		}
		
		// Resolve request via synchronous XHR
		try {
			var request = new XMLHttpRequest();		
			request.open('GET', chrome_path, false);
			request.send(null);
			if (request.status != 0) {
				return this.doError(500, "Error " + request.status + " retrieving '"+chrome_path+"'");
			}
			var page = request.responseText;
		} catch (ex) {
			return this.doError(500, "Error making page request");
		}
		
		// Parse page server-side templates if any exist
		if (path.toLowerCase().substr(path.length - 3, 3) == 'htm') {
			var right = page;
			var left = [];
		
			var i;
			for(;;) {
				var i = right.indexOf('<%');
				if (i == -1) {
					left.push(right);
					break;
				}
				left.push(right.substr(0, i));
				right = right.substr(i+2);
				
				var j = right.indexOf('%>');
				if (j == -1) {
					return this.doError(500, "Mismatched server-side script tags");
				}
				
				left.push( tb_eval(right.substr(0, j), q_params ) );
				right = right.substr(j+2);
			}
			page = left.join('');
			
		}
		
		// Return
		var http_response;
		http_response = 'HTTP/1.1 200 OK\nConnection: keep-alive\nContent-Length: '+page.length+"\n\n"+page;
		return http_response;
	}
}

//////////////////
// Eval context //
//////////////////

function tb_eval(code, params) {
	try {
		return eval(code);
	} catch (e) {
		return '<h1>' + e.name + ' error in server-side script: ' + e.message + '</h1>';
	}
}

/////////////////////
// Extension class //
/////////////////////

var thunderbirdwebui = {
	serverSocket : 0,

	onLoad: function() {
		this.initialized = true;
		this.restartWebListener();
	},
	
	restartWebListener: function() {
		this.serverSocket =
			Components.classes["@mozilla.org/network/server-socket;1"]
			.createInstance(Components.interfaces.nsIServerSocket)
		;
		this.serverSocket.init(8082, false, -1);
		this.serverSocket.asyncListen(tbw_listener);
	}
};

///////////////////////////
// Extension entry point //
///////////////////////////

window.addEventListener(
	"load",
	function (e) {
		thunderbirdwebui.onLoad();
	},
	false
);

